#!/usr/bin/env python3

"""
Tool for plotting a custom figure.
"""

import logging
from typing import Type, List, TypedDict, Annotated, Tuple, Union, Literal
from typing import Type, List, TypedDict, Annotated, Tuple, Union, Literal
from pydantic import BaseModel, Field
import pandas as pd
import pandas as pd
from langchain_openai import ChatOpenAI
from langchain_core.tools import BaseTool
from langgraph.prebuilt import InjectedState
from langgraph.prebuilt import InjectedState

# Initialize logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class CustomPlotterInput(BaseModel):
    """
    Input schema for the PlotImage tool.
    """
    question: str = Field(description="Description of the plot")
    state: Annotated[dict, InjectedState]
    state: Annotated[dict, InjectedState]

# Note: It's important that every field has type hints.
# BaseTool is a Pydantic class and not having type hints
# can lead to unexpected behavior.
# Note: It's important that every field has type hints.
# BaseTool is a Pydantic class and not having type hints
# can lead to unexpected behavior.
class CustomPlotterTool(BaseTool):
    """
    Tool for making custom plots
    """
    name: str = "custom_plotter"
    description: str = "A tool to make custom plots of the simulation results"
    args_schema: Type[BaseModel] = CustomPlotterInput
    response_format: str = "content_and_artifact"
    response_format: str = "content_and_artifact"

    def _run(self,
             question: str,
             state: Annotated[dict, InjectedState]
             ) -> Tuple[str, Union[None, List[str]]]:
        """
        Run the tool.

        Args:
            question (str): The question about the custom plot.
            state (dict): The state of the graph.
            question (str): The question about the custom plot.
            state (dict): The state of the graph.

        Returns:
            str: The answer to the question
        """
        logger.log(logging.INFO, "Calling custom_plotter tool %s", question)
        # Check if the simulation results are available
        # if 'dic_simulated_data' not in state:
        #     return "Please run the simulation first before plotting the figure.", None
        df = pd.DataFrame.from_dict(state['dic_simulated_data'])
        species_names = df.columns.tolist()
        # Exclude the time column
        species_names.remove('Time')
        # In the following code, we extract the species
        # from the user question. We use Literal to restrict
        # the species names to the ones available in the
        # simulation results.
        class CustomHeader(TypedDict):
            """
            A list of species based on user question.
            """
            relevant_species: Union[None, List[Literal[*species_names]]] = Field(
                    description="List of species based on user question. If no relevant species are found, it will be None.")
        # Create an instance of the LLM model
        llm = ChatOpenAI(model=state['llm_model'], temperature=0)
        llm_with_structured_output = llm.with_structured_output(CustomHeader)
        results = llm_with_structured_output.invoke(question)
        extracted_species = []
        # Extract the species from the results
        # that are available in the simulation results
        for species in results['relevant_species']:
            if species in species_names:
                extracted_species.append(species)
        logger.info("Extracted species: %s", extracted_species)
        if len(extracted_species) == 0:
            return "No species found in the simulation results that matches the user prompt.", None
        content = f"Plotted custom figure with species: {', '.join(extracted_species)}"
        return content, extracted_species



#!/usr/bin/env python3

"""
Tool for asking a question about the simulation results.
"""

import logging
from typing import Type, Annotated
import pandas as pd
from pydantic import BaseModel, Field
from langchain_core.tools.base import BaseTool
from langchain.agents.agent_types import AgentType
from langchain_experimental.agents import create_pandas_dataframe_agent
from langchain_openai import ChatOpenAI
from langgraph.prebuilt import InjectedState

# Initialize logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class AskQuestionInput(BaseModel):
    """
    Input schema for the AskQuestion tool.
    """
    question: str = Field(description="question about the simulation results")
    state: Annotated[dict, InjectedState]

# Note: It's important that every field has type hints.
# BaseTool is a Pydantic class and not having type hints
# can lead to unexpected behavior.
class AskQuestionTool(BaseTool):
    """
    Tool for calculating the product of two numbers.
    """
    name: str = "ask_question"
    description: str = "A tool to ask question about the simulation results."
    args_schema: Type[BaseModel] = AskQuestionInput
    return_direct: bool = False

    def _run(self,
             question: str,
             state: Annotated[dict, InjectedState]) -> str:
        """
        Run the tool.

        Args:
            question (str): The question to ask about the simulation results.
            state (dict): The state of the graph.
            run_manager (Optional[CallbackManagerForToolRun]): The CallbackManagerForToolRun object.

        Returns:
            str: The answer to the question.
        """
        logger.log(logging.INFO,
                   "Calling ask_question tool %s", question)
        # Check if the simulation results are available
        if 'dic_simulated_data' not in state:
            return "Please run the simulation first before \
                asking a question about the simulation results."
        df = pd.DataFrame.from_dict(state['dic_simulated_data'])
        prompt_content = None
        # if run_manager and 'prompt' in run_manager.metadata:
        #     prompt_content = run_manager.metadata['prompt']
        # Create a pandas dataframe agent with OpenAI
        df_agent = create_pandas_dataframe_agent(
                        ChatOpenAI(model=state['llm_model']),
                        allow_dangerous_code=True,
                        agent_type=AgentType.OPENAI_FUNCTIONS,
                        df=df,
                        prefix=prompt_content)
        llm_result = df_agent.invoke(question)
        return llm_result["output"]



#!/usr/bin/env python3

'''
This is the state file for the Talk2BioModels agent.
'''

from typing import Annotated
import operator
from langgraph.prebuilt.chat_agent_executor import AgentState

class Talk2Biomodels(AgentState):
    """
    The state for the Talk2BioModels agent.
    """
    model_id: Annotated[list, operator.add]
    # sbml_file_path: str
    # A StateGraph may receive a concurrent updates
    # which is not supported by the StateGraph.
    # Therefore, we need to use Annotated to specify
    # the operator for the sbml_file_path field.
    # https://langchain-ai.github.io/langgraph/troubleshooting/errors/INVALID_CONCURRENT_GRAPH_UPDATE/
    sbml_file_path: Annotated[list, operator.add]
    dic_simulated_data: dict
    llm_model: str



'''
This file is used to import all the modules in the package.
'''
from . import search_models
from . import simulate_model
from . import ask_question
from . import custom_plotter
from . import get_modelinfo
from . import load_biomodel



#!/usr/bin/env python3

"""
Tool for searching models based on search query.
"""

from typing import Type, Annotated
from pydantic import BaseModel, Field
from basico import biomodels
from langchain_core.tools import BaseTool
from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import ChatPromptTemplate
from langchain_openai import ChatOpenAI
from langgraph.prebuilt import InjectedState

class SearchModelsInput(BaseModel):
    """
    Input schema for the search models tool.
    """
    query: str = Field(description="Search models query", default=None)
    state: Annotated[dict, InjectedState]

# Note: It's important that every field has type hints. BaseTool is a
# Pydantic class and not having type hints can lead to unexpected behavior.
class SearchModelsTool(BaseTool):
    """
    Tool for returning the search results based on the search query.
    """
    name: str = "search_models"
    description: str = "Search models in the BioMmodels database based on keywords."
    args_schema: Type[BaseModel] = SearchModelsInput
    return_direct: bool = True

    def _run(self,
             query: str,
             state: Annotated[dict, InjectedState]) -> dict:
        """
        Run the tool.

        Args:
            query (str): The search query.

        Returns:
            dict: The answer to the question in the form of a dictionary.
        """
        search_results = biomodels.search_for_model(query)
        llm = ChatOpenAI(model=state['llm_model'])
        # Check if run_manager's metadata has the key 'prompt_content'
        prompt_content = f'''
                        Convert the input into a table.

                        The table must contain the following columns:
                        - #
                        - BioModel ID
                        - BioModel Name
                        - Format
                        - Submission Date

                        Additional Guidelines:
                        - The column # must contain the row number starting from 1.
                        - Embed the url for each BioModel ID in the table 
                        in the first column in the markdown format.
                        - The Submission Date must be in the format YYYY-MM-DD.

                        Input:
                        {input}.
                        '''
        prompt_template = ChatPromptTemplate.from_messages(
            [("system", prompt_content),
             ("user", "{input}")]
        )
        parser = StrOutputParser()
        chain = prompt_template | llm | parser
        return chain.invoke({"input": search_results})



"""
Class for loading StarkQAPrimeKG dataset.
"""

import os
import shutil
import pickle
import numpy as np
import pandas as pd
from tqdm import tqdm
import torch
from huggingface_hub import hf_hub_download, list_repo_files
import gdown
from .dataset import Dataset

class StarkQAPrimeKG(Dataset):
    """
    Class for loading StarkQAPrimeKG dataset.
    It downloads the data from the HuggingFace repo and stores it in the local directory.
    The data is then loaded into pandas DataFrame of QA pairs, dictionary of split indices,
    and node information.
    """

    def __init__(self, local_dir: str = "../../../data/starkqa_primekg/"):
        """
        Constructor for StarkQAPrimeKG class.

        Args:
            local_dir (str): The local directory to store the dataset files.
        """
        self.name: str = "starkqa_primekg"
        self.hf_repo_id: str = "snap-stanford/stark"
        self.local_dir: str = local_dir
        # Attributes to store the data
        self.starkqa: pd.DataFrame = None
        self.starkqa_split_idx: dict = None
        self.starkqa_node_info: dict = None
        self.query_emb_dict: dict = None
        self.node_emb_dict: dict = None

        # Set up the dataset
        self.setup()

    def setup(self):
        """
        A method to set up the dataset.
        """
        # Make the directory if it doesn't exist
        os.makedirs(os.path.dirname(self.local_dir), exist_ok=True)

    def _load_stark_repo(self) -> tuple[pd.DataFrame, dict, dict]:
        """
        Private method to load related files of StarkQAPrimeKG dataset.

        Returns:
            The nodes dataframe of StarkQAPrimeKG dataset.
            The split indices of StarkQAPrimeKG dataset.
            The node information of StarkQAPrimeKG dataset.
        """
        # Download the file if it does not exist in the local directory
        # Otherwise, load the data from the local directory
        local_file = os.path.join(self.local_dir, "qa/prime/stark_qa/stark_qa.csv")
        if os.path.exists(local_file):
            print(f"{local_file} already exists. Loading the data from the local directory.")
        else:
            print(f"Downloading files from {self.hf_repo_id}")

            # List all related files in the HuggingFace Hub repository
            files = list_repo_files(self.hf_repo_id, repo_type="dataset")
            files = [f for f in files if ((f.startswith("qa/prime/") or
                                           f.startswith("skb/prime/")) and f.find("raw") == -1)]

            # Download and save each file in the specified folder
            for file in tqdm(files):
                _ = hf_hub_download(self.hf_repo_id,
                                    file,
                                    repo_type="dataset",
                                    local_dir=self.local_dir)

            # Unzip the processed files
            shutil.unpack_archive(
                os.path.join(self.local_dir, "skb/prime/processed.zip"),
                os.path.join(self.local_dir, "skb/prime/")
            )

        # Load StarkQA dataframe
        starkqa = pd.read_csv(
            os.path.join(self.local_dir, "qa/prime/stark_qa/stark_qa.csv"),
            low_memory=False)

        # Read split indices
        qa_indices = sorted(starkqa['id'].tolist())
        starkqa_split_idx = {}
        for split in ['train', 'val', 'test', 'test-0.1']:
            indices_file = os.path.join(self.local_dir, "qa/prime/split", f'{split}.index')
            with open(indices_file, 'r', encoding='utf-8') as f:
                indices = f.read().strip().split('\n')
            query_ids = [int(idx) for idx in indices]
            starkqa_split_idx[split] = np.array(
                [qa_indices.index(query_id) for query_id in query_ids]
            )

        # Load the node info of PrimeKG preprocessed for StarkQA
        with open(os.path.join(self.local_dir, 'skb/prime/processed/node_info.pkl'), 'rb') as f:
            starkqa_node_info = pickle.load(f)

        return starkqa, starkqa_split_idx, starkqa_node_info

    def _load_stark_embeddings(self) -> tuple[dict, dict]:
        """
        Private method to load the embeddings of StarkQAPrimeKG dataset.

        Returns:
            The query embeddings of StarkQAPrimeKG dataset.
            The node embeddings of StarkQAPrimeKG dataset.
        """
        # Load the provided embeddings of query and nodes
        # Note that they utilized 'text-embedding-ada-002' for embeddings
        emb_model = 'text-embedding-ada-002'
        query_emb_url = 'https://drive.google.com/uc?id=1MshwJttPZsHEM2cKA5T13SIrsLeBEdyU'
        node_emb_url = 'https://drive.google.com/uc?id=16EJvCMbgkVrQ0BuIBvLBp-BYPaye-Edy'

        # Prepare respective directories to store the embeddings
        emb_dir = os.path.join(self.local_dir, emb_model)
        query_emb_dir = os.path.join(emb_dir, "query")
        node_emb_dir = os.path.join(emb_dir, "doc")
        os.makedirs(query_emb_dir, exist_ok=True)
        os.makedirs(node_emb_dir, exist_ok=True)
        query_emb_path = os.path.join(query_emb_dir, "query_emb_dict.pt")
        node_emb_path = os.path.join(node_emb_dir, "candidate_emb_dict.pt")

        # Download the embeddings if they do not exist in the local directory
        if not os.path.exists(query_emb_path) or not os.path.exists(node_emb_path):
            # Download the query embeddings
            gdown.download(query_emb_url, query_emb_path, quiet=False)

            # Download the node embeddings
            gdown.download(node_emb_url, node_emb_path, quiet=False)

        # Load the embeddings
        query_emb_dict = torch.load(query_emb_path)
        node_emb_dict = torch.load(node_emb_path)

        return query_emb_dict, node_emb_dict

    def load_data(self):
        """
        Load the StarkQAPrimeKG dataset into pandas DataFrame of QA pairs,
        dictionary of split indices, and node information.
        """
        print("Loading StarkQAPrimeKG dataset...")
        self.starkqa, self.starkqa_split_idx, self.starkqa_node_info = self._load_stark_repo()

        print("Loading StarkQAPrimeKG embeddings...")
        self.query_emb_dict, self.node_emb_dict = self._load_stark_embeddings()


    def get_starkqa(self) -> pd.DataFrame:
        """
        Get the dataframe of StarkQAPrimeKG dataset, containing the QA pairs.

        Returns:
            The nodes dataframe of PrimeKG dataset.
        """
        return self.starkqa

    def get_starkqa_split_indicies(self) -> dict:
        """
        Get the split indices of StarkQAPrimeKG dataset.

        Returns:
            The split indices of StarkQAPrimeKG dataset.
        """
        return self.starkqa_split_idx

    def get_starkqa_node_info(self) -> dict:
        """
        Get the node information of StarkQAPrimeKG dataset.

        Returns:
            The node information of StarkQAPrimeKG dataset.
        """
        return self.starkqa_node_info

    def get_query_embeddings(self) -> dict:
        """
        Get the query embeddings of StarkQAPrimeKG dataset.

        Returns:
            The query embeddings of StarkQAPrimeKG dataset.
        """
        return self.query_emb_dict

    def get_node_embeddings(self) -> dict:
        """
        Get the node embeddings of StarkQAPrimeKG dataset.

        Returns:
            The node embeddings of StarkQAPrimeKG dataset.
        """
        return self.node_emb_dict



#!/usr/bin/env python3

"""
Tool for simulating a model.
"""

import logging
from dataclasses import dataclass
from typing import Type, Union, List, Annotated
import basico
from pydantic import BaseModel, Field
from langgraph.types import Command
from langgraph.prebuilt import InjectedState
from langchain_core.tools import BaseTool
from langchain_core.messages import ToolMessage
from langchain_core.tools.base import InjectedToolCallId
from .load_biomodel import ModelData, load_biomodel

# Initialize logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

@dataclass
class TimeData:
    """
    Dataclass for storing the time data.
    """
    duration: Union[int, float] = 100
    interval: Union[int, float] = 10

@dataclass
class SpeciesData:
    """
    Dataclass for storing the species data.
    """
    species_name: List[str] = Field(description="species name", default=None)
    species_concentration: List[Union[int, float]] = Field(
        description="initial species concentration",
        default=None)

@dataclass
class TimeSpeciesNameConcentration:
    """
    Dataclass for storing the time, species name, and concentration data.
    """
    time: Union[int, float] = Field(description="time point where the event occurs")
    species_name: str = Field(description="species name")
    species_concentration: Union[int, float] = Field(
        description="species concentration at the time point")

@dataclass
class RecurringData:
    """
    Dataclass for storing the species and time data 
    on recurring basis.
    """
    data: List[TimeSpeciesNameConcentration] = Field(
        description="species and time data on recurring basis",
        default=None)

@dataclass
class ArgumentData:
    """
    Dataclass for storing the argument data.
    """
    time_data: TimeData = Field(description="time data", default=None)
    species_data: SpeciesData = Field(
        description="species name and initial concentration data",
        default=None)
    recurring_data: RecurringData = Field(
        description="species and time data on recurring basis",
        default=None)

def add_rec_events(model_object, recurring_data):
    """
    Add recurring events to the model.
    """
    for row in recurring_data.data:
        tp, sn, sc = row.time, row.species_name, row.species_concentration
        basico.add_event(f'{sn}_{tp}',
                            f'Time > {tp}',
                            [[sn, str(sc)]],
                            model=model_object.copasi_model)

class SimulateModelInput(BaseModel):
    """
    Input schema for the SimulateModel tool.
    """
    sys_bio_model: ModelData = Field(description="model data", default=None)
    arg_data: ArgumentData = Field(description="time, species, and recurring data",
                                        default=None)
    tool_call_id: Annotated[str, InjectedToolCallId]
    state: Annotated[dict, InjectedState]

# Note: It's important that every field has type hints. BaseTool is a
# Pydantic class and not having type hints can lead to unexpected behavior.
class SimulateModelTool(BaseTool):
    """
    Tool for simulating a model.
    """
    name: str = "simulate_model"
    description: str = "A tool to simulate a biomodel"
    args_schema: Type[BaseModel] = SimulateModelInput

    def _run(self,
        tool_call_id: Annotated[str, InjectedToolCallId],
        state: Annotated[dict, InjectedState],
        sys_bio_model: ModelData = None,
        arg_data: ArgumentData = None
    ) -> Command:
        """
        Run the tool.

        Args:
            tool_call_id (str): The tool call ID. This is injected by the system.
            state (dict): The state of the tool.
            sys_bio_model (ModelData): The model data.
            arg_data (ArgumentData): The argument data.

        Returns:
            str: The result of the simulation.
        """
        logger.log(logging.INFO,
                   "Calling simulate_model tool %s, %s",
                   sys_bio_model,
                   arg_data)
        sbml_file_path = state['sbml_file_path'][-1] if len(state['sbml_file_path']) > 0 else None
        model_object = load_biomodel(sys_bio_model,
                                  sbml_file_path=sbml_file_path)
        # Prepare the dictionary of species data
        # that will be passed to the simulate method
        # of the BasicoModel class
        duration = 100.0
        interval = 10
        dic_species_data = None
        if arg_data:
            # Prepare the dictionary of species data
            if arg_data.species_data is not None:
                dic_species_data = dict(zip(arg_data.species_data.species_name,
                                            arg_data.species_data.species_concentration))
            # Add recurring events (if any) to the model
            if arg_data.recurring_data is not None:
                add_rec_events(model_object, arg_data.recurring_data)
            # Set the duration and interval
            if arg_data.time_data is not None:
                duration = arg_data.time_data.duration
                interval = arg_data.time_data.interval

        # Simulate the model
        df = model_object.simulate(
            parameters=dic_species_data,
            duration=duration,
            interval=interval
            )

        # Prepare the dictionary of updated state for the model
        dic_updated_state_for_model = {}
        for key, value in {
                        "model_id": [sys_bio_model.biomodel_id],
                        "sbml_file_path": [sbml_file_path],
                        }.items():
            if value:
                dic_updated_state_for_model[key] = value

        # Return the updated state of the tool
        return Command(
                update=dic_updated_state_for_model|{
                # update the state keys
                "dic_simulated_data": df.to_dict(),
                # update the message history
                "messages": [
                    ToolMessage(
                        content="Simulation results are ready.",
                        tool_call_id=tool_call_id
                        )
                    ],
                }
            )



#!/usr/bin/env python3

"""
Function for loading the BioModel.
"""

from dataclasses import dataclass
from ..models.basico_model import BasicoModel

@dataclass
class ModelData:
    """
    Dataclass for storing the model data.
    """
    biomodel_id: int = None
    # sbml_file_path: Optional[str] = None
    use_uploaded_sbml_file: bool = False

def load_biomodel(sys_bio_model, sbml_file_path=None):
    """
    Load the BioModel.
    """
    model_object = None
    if sys_bio_model.biomodel_id:
        model_object = BasicoModel(biomodel_id=sys_bio_model.biomodel_id)
    elif sbml_file_path:
        model_object = BasicoModel(sbml_file_path=sbml_file_path)
    return model_object



"""
Class for loading BioBridgePrimeKG dataset.
"""

import os
import pickle
import json
import requests
import numpy as np
import pandas as pd
from tqdm import tqdm
from .dataset import Dataset
from .primekg import PrimeKG

class BioBridgePrimeKG(Dataset):
    """
    Class for loading BioBridgePrimeKG dataset.
    It downloads the data from the BioBridge repo and stores it in the local directory.
    The data is then loaded into pandas DataFrame of nodes and edges.
    This class was adapted from the BioBridge repo:
    https://github.com/RyanWangZf/BioBridge
    """

    def __init__(self,
                 primekg_dir: str = "../../../data/primekg/",
                 local_dir: str = "../../../data/biobridge_primekg/",
                 random_seed: int=0,
                 n_neg_samples: int=5):
        """
        Constructor for BioBridgePrimeKG class.

        Args:
            primekg_dir (str): The directory of PrimeKG dataset.
            local_dir (str): The directory to store the downloaded data.
            random_seed (int): The random seed value.
        """
        self.name: str = "biobridge_primekg"
        self.primekg_dir: str = primekg_dir
        self.local_dir: str = local_dir
        self.random_seed = random_seed
        self.n_neg_samples = n_neg_samples
        # Preselected node types:
        # protein, molecular function, cellular component, biological process, drug, disease
        self.preselected_node_types = ["protein", "mf", "cc", "bp", "drug", "disease"]
        self.node_type_map = {
            "protein": "gene/protein",
            "mf": "molecular_function",
            "cc": "cellular_component",
            "bp": "biological_process",
            "drug": "drug",
            "disease": "disease",
        }

        # Attributes to store the data
        self.primekg = None
        self.primekg_triplets = None
        self.primekg_triplets_negative = None
        self.data_config = None
        self.emb_dict = None
        self.df_train = None
        self.df_node_train = None
        self.df_test = None
        self.df_node_test = None
        self.node_info_dict = None

        # Set up the dataset
        self.setup()

    def setup(self):
        """
        A method to set up the dataset.
        """
        # Make the directories if it doesn't exist
        os.makedirs(os.path.dirname(self.primekg_dir), exist_ok=True)
        os.makedirs(os.path.dirname(self.local_dir), exist_ok=True)

        # Set the random seed
        self.set_random_seed(self.random_seed)

        # Set SettingWithCopyWarning  warnings to none
        pd.options.mode.chained_assignment = None

    def _load_primekg(self) -> PrimeKG:
        """
        Private method to load related files of PrimeKG dataset.

        Returns:
            The PrimeKG dataset.
        """
        primekg_data = PrimeKG(local_dir=self.primekg_dir)
        primekg_data.load_data()

        return primekg_data

    def _download_file(self,
                       remote_url:str,
                       local_dir: str,
                       local_filename: str):
        """
        A helper function to download a file from remote URL to the local directory.

        Args:
            remote_url (str): The remote URL of the file to be downloaded.
            local_dir (str): The local directory to store the downloaded file.
            local_filename (str): The local filename to store the downloaded file.
        """
        # Make the local directory if it does not exist
        if not os.path.exists(local_dir):
            os.makedirs(local_dir)
        # Download the file from remote URL to local directory
        local_path = os.path.join(local_dir, local_filename)
        if os.path.exists(local_path):
            print(f"File {local_filename} already exists in {local_dir}.")
        else:
            print(f"Downloading {local_filename} from {remote_url} to {local_dir}...")
            response = requests.get(remote_url, stream=True, timeout=300)
            response.raise_for_status()
            progress_bar = tqdm(
                total=int(response.headers.get("content-length", 0)),
                unit="iB",
                unit_scale=True,
            )
            with open(os.path.join(local_dir, local_filename), "wb") as file:
                for data in response.iter_content(1024):
                    progress_bar.update(len(data))
                    file.write(data)
            progress_bar.close()

    def _load_data_config(self) -> dict:
        """
        Load the data config file of BioBridgePrimeKG dataset.

        Returns:
            The data config file of BioBridgePrimeKG dataset.
        """
        # Download the data config file of BioBridgePrimeKG
        self._download_file(
            remote_url= ('https://raw.githubusercontent.com/RyanWangZf/BioBridge/'
                         'refs/heads/main/data/BindData/data_config.json'),
            local_dir=self.local_dir,
            local_filename='data_config.json')

        # Load the downloaded data config file
        with open(os.path.join(self.local_dir, 'data_config.json'), 'r', encoding='utf-8') as f:
            data_config = json.load(f)

        return data_config

    def _build_node_embeddings(self) -> dict:
        """
        Build the node embeddings for BioBridgePrimeKG dataset.

        Returns:
            The dictionary of node embeddings.
        """
        processed_file_path = os.path.join(self.local_dir, "embeddings", "embedding_dict.pkl")
        if os.path.exists(processed_file_path):
            # Load the embeddings from the local directory
            with open(processed_file_path, "rb") as f:
                emb_dict_all = pickle.load(f)
        else:
            # Download the embeddings from the BioBridge repo and further process them
            # List of embedding source files
            url = ('https://media.githubusercontent.com/media/RyanWangZf/BioBridge/'
                   'refs/heads/main/data/embeddings/esm2b_unimo_pubmedbert/')
            file_list = [f"{n}.pkl" for n in self.preselected_node_types]

            # Download the embeddings
            for file in file_list:
                self._download_file(remote_url=os.path.join(url, file),
                                    local_dir=os.path.join(self.local_dir, "embeddings"),
                                    local_filename=file)

            # Unified embeddings
            emb_dict_all = {}
            for file in file_list:
                with open(os.path.join(self.local_dir, "embeddings", file), "rb") as f:
                    emb = pickle.load(f)
                emb_ar = emb["embedding"]
                if not isinstance(emb_ar, list):
                    emb_ar = emb_ar.tolist()
                emb_dict_all.update(dict(zip(emb["node_index"], emb_ar)))

            # Store embeddings
            with open(processed_file_path, "wb") as f:
                pickle.dump(emb_dict_all, f)

        return emb_dict_all

    def _build_full_triplets(self) -> tuple[pd.DataFrame, dict]:
        """
        Build the full triplets for BioBridgePrimeKG dataset.

        Returns:
            The full triplets for BioBridgePrimeKG dataset.
            The dictionary of node information.
        """
        processed_file_path = os.path.join(self.local_dir, "processed", "triplet_full.tsv.gz")
        if os.path.exists(processed_file_path):
            # Load the file from the local directory
            with open(processed_file_path, "rb") as f:
                primekg_triplets = pd.read_csv(f, sep="\t", compression="gzip", low_memory=False)

            # Load each dataframe in the local directory
            node_info_dict = {}
            for i, node_type in enumerate(self.preselected_node_types):
                with open(os.path.join(self.local_dir, "processed",
                                       f"{node_type}.csv"), "rb") as f:
                    df_node = pd.read_csv(f)
                node_info_dict[self.node_type_map[node_type]] = df_node
        else:
            # Download the related files from the BioBridge repo and further process them
            # List of processed files
            url = ('https://media.githubusercontent.com/media/RyanWangZf/BioBridge/'
                   'refs/heads/main/data/Processed/')
            file_list = ["protein", "molecular", "cellular", "biological", "drug", "disease"]

            # Download the processed files
            for i, file in enumerate(file_list):
                self._download_file(remote_url=os.path.join(url, f"{file}.csv"),
                                    local_dir=os.path.join(self.local_dir, "processed"),
                                    local_filename=f"{self.preselected_node_types[i]}.csv")

            # Build the node index list
            node_info_dict = {}
            node_index_list = []
            for i, file in enumerate(file_list):
                df_node = pd.read_csv(os.path.join(self.local_dir, "processed",
                                                   f"{self.preselected_node_types[i]}.csv"))
                node_info_dict[self.node_type_map[self.preselected_node_types[i]]] = df_node
                node_index_list.extend(df_node["node_index"].tolist())

            # Filter the PrimeKG dataset to take into account only the selected node types
            primekg_triplets = self.primekg.get_edges().copy()
            primekg_triplets = primekg_triplets[
                primekg_triplets["head_index"].isin(node_index_list) &\
                primekg_triplets["tail_index"].isin(node_index_list)
            ]
            primekg_triplets = primekg_triplets.reset_index(drop=True)

            # Perform mapping of node types
            primekg_triplets["head_type"] = primekg_triplets["head_type"].apply(
                lambda x: self.data_config["node_type"][x]
            )
            primekg_triplets["tail_type"] = primekg_triplets["tail_type"].apply(
                lambda x: self.data_config["node_type"][x]
            )

            # Perform mapping of relation types
            primekg_triplets["display_relation"] = primekg_triplets["display_relation"].apply(
                lambda x: self.data_config["relation_type"][x]
            )

            # Store the processed triplets
            primekg_triplets.to_csv(processed_file_path, sep="\t", compression="gzip", index=False)

        return primekg_triplets, node_info_dict

    def _build_train_test_split(self) -> tuple[pd.DataFrame, pd.DataFrame,
                                               pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """
        Build the train-test split for BioBridgePrimeKG dataset.

        Returns:
            The train triplets for BioBridgePrimeKG dataset.
            The train nodes for BioBridgePrimeKG dataset.
            The test triplets for BioBridgePrimeKG dataset.
            The test nodes for BioBridgePrimeKG dataset.
            The full triplets for BioBridgePrimeKG dataset.
        """
        if os.path.exists(os.path.join(self.local_dir, "processed",
                                       "triplet_full_altered.tsv.gz")):
            # Load each dataframe in the local directory
            with open(os.path.join(self.local_dir, "processed",
                                   "triplet_train.tsv.gz"), "rb") as f:
                df_train = pd.read_csv(f, sep="\t", compression="gzip", low_memory=False)

            with open(os.path.join(self.local_dir, "processed",
                                   "node_train.tsv.gz"), "rb") as f:
                df_node_train = pd.read_csv(f, sep="\t", compression="gzip", low_memory=False)

            with open(os.path.join(self.local_dir, "processed",
                                   "triplet_test.tsv.gz"), "rb") as f:
                df_test = pd.read_csv(f, sep="\t", compression="gzip", low_memory=False)

            with open(os.path.join(self.local_dir, "processed",
                                   "node_test.tsv.gz"), "rb") as f:
                df_node_test = pd.read_csv(f, sep="\t", compression="gzip", low_memory=False)

            with open(os.path.join(self.local_dir, "processed",
                                   "triplet_full_altered.tsv.gz"), "rb") as f:
                triplets = pd.read_csv(f, sep="\t", compression="gzip", low_memory=False)
        else:
            # Filtering out some nodes in the embedding dictionary
            triplets = self.primekg_triplets.copy()
            triplets = triplets[
                triplets["head_index"].isin(list(self.emb_dict.keys())) &\
                triplets["tail_index"].isin(list(self.emb_dict.keys()))
            ].reset_index(drop=True)

            # Perform splitting of the triplets
            list_split = {
                "train": [],
                "test": [],
            }
            node_split = {
                "train": {
                    "node_index": [],
                    "node_type": [],
                },
                "test": {
                    "node_index": [],
                    "node_type": [],
                }
            }
            # Loop over the node types
            for node_type in triplets["head_type"].unique():
                df_sub = triplets[triplets["head_type"] == node_type]
                all_x_indexes = df_sub["head_index"].unique()
                # By default, we use 90% of the nodes for training and 10% for testing
                te_x_indexes = np.random.choice(
                    all_x_indexes, size=int(0.1*len(all_x_indexes)), replace=False
                )
                df_subs = {}
                df_subs["test"] = df_sub[df_sub["head_index"].isin(te_x_indexes)]
                df_subs["train"] = df_sub[~df_sub["head_index"].isin(te_x_indexes)]
                list_split["train"].append(df_subs["train"])
                list_split["test"].append(df_subs["test"])

                # record the split
                node_index = {}
                node_index["train"] = df_subs["train"]["head_index"].unique()
                node_split["train"]["node_index"].extend(node_index["train"].tolist())
                node_split["train"]["node_type"].extend([node_type]*len(node_index["train"]))
                node_index["test"] = df_subs["test"]["head_index"].unique()
                node_split["test"]["node_index"].extend(node_index["test"].tolist())
                node_split["test"]["node_type"].extend([node_type]*len(node_index["test"]))

                print(f"Number of {node_type} nodes in train: {len(node_index['train'])}")
                print(f"Number of {node_type} nodes in test: {len(node_index['test'])}")

            # Prepare train and test DataFrames
            df_train = pd.concat(list_split["train"])
            df_node_train = pd.DataFrame(node_split["train"])
            df_test = pd.concat(list_split["test"])
            df_node_test = pd.DataFrame(node_split["test"])

            # Store each dataframe in the local directory
            df_train.to_csv(os.path.join(self.local_dir, "processed", "triplet_train.tsv.gz"),
                            sep="\t", compression="gzip", index=False)
            df_node_train.to_csv(os.path.join(self.local_dir, "processed", "node_train.tsv.gz"),
                                sep="\t", compression="gzip", index=False)
            df_test.to_csv(os.path.join(self.local_dir, "processed", "triplet_test.tsv.gz"),
                           sep="\t", compression="gzip", index=False)
            df_node_test.to_csv(os.path.join(self.local_dir, "processed", "node_test.tsv.gz"),
                                sep="\t", compression="gzip", index=False)
            # Store altered full triplets as well
            triplets.to_csv(os.path.join(self.local_dir, "processed",
                                         "triplet_full_altered.tsv.gz"),
                            sep="\t", compression="gzip", index=False)

        return df_train, df_node_train, df_test, df_node_test, triplets

    # def _negative_sampling(self,
    #                        batch_df: pd.DataFrame,
    #                        process_index: int,
    #                        index_map: dict,
    #                        node_train_dict: dict) -> pd.DataFrame:
    #     """
    #     A helper function to perform negative sampling for a batch of triplets.
    #     """
    #     negative_y_index_list = []
    #     for _, row in tqdm(batch_df.iterrows(),
    #                        total=batch_df.shape[0],
    #                        desc=f"Process {process_index}"):
    #         x_index = row['head_index']
    #         # y_index = row['y_index']
    #         y_index_type = row['tail_type']
    #         paired_y_index_list = index_map[x_index]

    #         # sample a list of negative y_index
    #         node_train_sub = node_train_dict[y_index_type]
    #         negative_y_index = node_train_sub[
    #             ~node_train_sub['node_index'].isin(paired_y_index_list)
    #         ]['node_index'].sample(self.n_neg_samples).tolist()
    #         negative_y_index_list.append(negative_y_index)

    #     batch_df.loc[:, 'negative_tail_index'] = negative_y_index_list
    #     return batch_df

    # def _build_negative_triplets(self,
    #                              chunk_size: int=100000,
    #                              n_neg_samples: int=10):
    #     """
    #     Build the negative triplets for BioBridgePrimeKG dataset.
    #     """
    #     processed_file_path = os.path.join(self.local_dir,
    #                                        "processed",
    #                                        "triplet_train_negative.tsv.gz")
    #     if os.path.exists(processed_file_path):
    #         # Load the negative triplets from the local directory
    #         with open(processed_file_path, "rb") as f:
    #             triplets_negative = pd.read_csv(f, sep="\t", compression="gzip", low_memory=False)
    #     else:
    #         # Set the number samples for negative sampling
    #         self.n_neg_samples = n_neg_samples

    #         # Split node list by type
    #         node_train_dict = {}
    #         type_list = self.df_node_train['node_type'].unique()
    #         for node_type in type_list:
    #             node_train_dict[node_type] = self.df_node_train[
    #                 self.df_node_train['node_type'] == node_type
    #             ].reset_index(drop=True)

    #         # create an index mapping from x_index to y_index
    #         index_map = self.df_train[
    #             ['head_index', 'tail_index']
    #         ].drop_duplicates().groupby('head_index').agg(list).to_dict()['tail_index']

    #         # Negative sampling
    #         batch_df_list = []
    #         for i in tqdm(range(0, self.df_train.shape[0], chunk_size)):
    #             batch_df_list.append(self.df_train.iloc[i:i+chunk_size])
    #         # Process negative sampling
    #         results = [
    #             self._negative_sampling(batch_df,
    #                                     num_piece,
    #                                     index_map,
    #                                     node_train_dict)
    #                                     for num_piece, batch_df in enumerate(batch_df_list)
    #         ]

    #         # Store the negative triplets
    #         triplets_negative = pd.concat(results, axis=0)
    #         triplets_negative.to_csv(processed_file_path,
    #                                  sep="\t", compression="gzip", index=False)

    #     # Set attribute
    #     self.primekg_triplets_negative = triplets_negative

    #     return triplets_negative

    # def load_data(self,
    #               build_neg_triplest: bool= False,
    #               chunk_size: int=100000,
    #               n_neg_samples: int=10):

    def load_data(self):
        """
        Load the BioBridgePrimeKG dataset into pandas DataFrame of nodes and edges.

        Args:
            build_neg_triplest (bool): Whether to build negative triplets.
            chunk_size (int): The chunk size for negative sampling.
            n_neg_samples (int): The number of negative samples for negative sampling.
        """
        # Load PrimeKG dataset
        print("Loading PrimeKG dataset...")
        self.primekg = self._load_primekg()

        # Load data config file of BioBridgePrimeKG
        print("Loading data config file of BioBridgePrimeKG...")
        self.data_config = self._load_data_config()

        # Build node embeddings
        print("Building node embeddings...")
        self.emb_dict = self._build_node_embeddings()

        # Build full triplets
        print("Building full triplets...")
        self.primekg_triplets, self.node_info_dict = self._build_full_triplets()

        # Build train-test split
        print("Building train-test split...")
        self.df_train, self.df_node_train, self.df_test, self.df_node_test, self.primekg_triplets =\
        self._build_train_test_split()

        # if build_neg_triplest:
        #     # Build negative triplets
        #     print("Building negative triplets...")
        #     self.primekg_triplets_negative = self._build_negative_triplets(
        #         chunk_size=chunk_size,
        #         n_neg_samples=n_neg_samples
        #     )

    def set_random_seed(self, seed: int):
        """
        Set the random seed for reproducibility.

        Args:
            seed (int): The random seed value.
        """
        np.random.seed(seed)

    def get_primekg(self) -> PrimeKG:
        """
        Get the PrimeKG dataset.

        Returns:
            The PrimeKG dataset.
        """
        return self.primekg

    def get_data_config(self) -> dict:
        """
        Get the data config file of BioBridgePrimeKG dataset.

        Returns:
            The data config file of BioBridgePrimeKG dataset.
        """
        return self.data_config

    def get_node_embeddings(self) -> dict:
        """
        Get the node embeddings for BioBridgePrimeKG dataset.

        Returns:
            The dictionary of node embeddings.
        """
        return self.emb_dict

    def get_primekg_triplets(self) -> pd.DataFrame:
        """
        Get the full triplets for BioBridgePrimeKG dataset.

        Returns:
            The full triplets for BioBridgePrimeKG dataset.
        """
        return self.primekg_triplets

    # def get_primekg_triplets_negative(self) -> pd.DataFrame:
    #     """
    #     Get the negative triplets for BioBridgePrimeKG dataset.

    #     Returns:
    #         The negative triplets for BioBridgePrimeKG dataset.
    #     """
    #     return self.primekg_triplets_negative

    def get_train_test_split(self) -> dict:
        """
        Get the train-test split for BioBridgePrimeKG dataset.

        Returns:
            The train-test split for BioBridgePrimeKG dataset.
        """
        return {
            "train": self.df_train,
            "node_train": self.df_node_train,
            "test": self.df_test,
            "node_test": self.df_node_test
        }

    def get_node_info_dict(self) -> dict:
        """
        Get the node information dictionary for BioBridgePrimeKG dataset.

        Returns:
            The node information dictionary for BioBridgePrimeKG dataset.
        """
        return self.node_info_dict
